import time

import bittensor as bt
import structlog
from django.conf import settings

import abstract_block_dumper.dal.django_dal as abd_dal
import abstract_block_dumper.services.utils as abd_utils
from abstract_block_dumper.services.block_processor import BlockProcessor, block_processor_factory

logger = structlog.get_logger(__name__)


class TaskScheduler:
    def __init__(
        self,
        block_processor: BlockProcessor,
        subtensor: bt.Subtensor,
        poll_interval: int,
    ) -> None:
        self.block_processor = block_processor
        self.subtensor = subtensor
        self.poll_interval = poll_interval
        self.last_processed_block = -1
        self.is_running = False

    def start(self) -> None:
        self.is_running = True

        self.initialize_last_block()

        logger.info(
            "TaskScheduler started",
            last_processed_block=self.last_processed_block,
            registry_functions=len(self.block_processor.registry.get_functions()),
        )

        while self.is_running:
            try:
                # Process lost retries first
                self.block_processor.recover_failed_retries()

                current_block = self.subtensor.get_current_block()

                for block_number in range(self.last_processed_block + 1, current_block + 1):
                    self.block_processor.process_block(block_number)
                    self.last_processed_block = block_number

                time.sleep(self.poll_interval)
            except KeyboardInterrupt:
                logger.info("TaskScheduler stopping due to KeyboardInterrupt.")
                self.stop()
                break
            except Exception:
                logger.error("Fatal scheduler error", exc_info=True)
                # resume the loop even if task failed
                time.sleep(self.poll_interval)

    def stop(self) -> None:
        self.is_running = False
        logger.info("TaskScheduler stopped.")

    def initialize_last_block(self) -> None:
        start_from_block_setting = getattr(settings, "BLOCK_DUMPER_START_FROM_BLOCK")

        if start_from_block_setting is not None:
            if start_from_block_setting == "current":
                self.last_processed_block = self.subtensor.get_current_block()
                logger.info(f"Starting from current blockchain block {self.last_processed_block}")

            elif isinstance(start_from_block_setting, int):
                self.last_processed_block = start_from_block_setting
                logger.info(f"Starting from configured block {self.last_processed_block}")
            else:
                raise ValueError(f"Invalid BLOCK_DUMPER_START_FROM_BLOCK value: {start_from_block_setting}")
        else:
            # Default behavior - resume from database
            last_block_number = abd_dal.get_the_latest_executed_block_number()

            self.last_processed_block = last_block_number or self.subtensor.get_current_block()
            logger.info(
                "Resume from the last database block or start from the current block",
                last_processed_block=self.last_processed_block,
            )


def task_scheduler_factory() -> TaskScheduler:
    return TaskScheduler(
        block_processor=block_processor_factory(),
        subtensor=abd_utils.get_bittensor_client(),
        poll_interval=getattr(settings, "BLOCK_DUMPER_POLL_INTERVAL", 1),
    )



from abstract_block_dumper.decorators import block_task


@block_task(
    condition=lambda bn: True,
)
def process_every_block(block_number: int, netuid: int | None = None):
    return f"Processed block {block_number} for NetUID {netuid}"


@block_task(
    condition=lambda bn: True,
    backfilling_lookback=100,
)
def backfill_previous_100_blocks(block_number: int, netuid: int | None = None):
    return f"Processed block {block_number} for NetUID {netuid}"


@block_task(
    condition=lambda bn, netuid: (bn + netuid) % 50 == 0,
    args=[{"netuid": i} for i in range(10, 15)],  # All subnets
    backfilling_lookback=1000,
    celery_kwargs={"retry": True},
)
def subnet_analysis(block_number, netuid):
    # Analyze subnet data
    return f"Processed block {block_number} for NetUID {netuid}"



import os

from celery import Celery
from django.conf import settings

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "example_project.settings")

app = Celery("example_project")

app.config_from_object(settings, namespace="CELERY")

app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)



import abc
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any

import structlog
from celery import Task

from abstract_block_dumper.exceptions import ConditionEvaluationError

logger = structlog.getLogger(__name__)


@dataclass
class RegistryItem:
    condition: Callable[..., bool]
    function: Task
    args: list[dict[str, Any]] | None = None
    backfilling_lookback: int | None = None
    celery_kwargs: dict[str, Any] = field(default_factory=dict)

    def match_condition(self, block_number: int, **kwargs) -> bool:
        """
        Check if condition matches for given block and arguments
        """
        try:
            return self.condition(block_number, **kwargs)
        except Exception as e:
            logger.error(
                "Condition evaluation failed",
                condition=self.function.__name__,
                block_number=block_number,
                exc_info=True,
            )
            raise ConditionEvaluationError(f"Failed to evaluate condition: {e}") from e

    def get_execution_args(self) -> list[dict[str, Any]]:
        """
        Get list of argument sets for execution
        """
        return self.args or [{}]

    @property
    def executable_path(self) -> str:
        """
        Get the importable path to the function.
        """
        if hasattr(self.function, "name") and self.function.name is not None:
            return self.function.name

        return ".".join([self.function.__module__, self.function.__name__])

    def requires_backfilling(self) -> bool:
        """
        Check if this item requires backfilling.
        """
        return self.backfilling_lookback is not None


class BaseRegistry(abc.ABC):
    @abc.abstractmethod
    def register_item(self, item: RegistryItem) -> None:
        pass

    @abc.abstractmethod
    def get_functions(self) -> list[RegistryItem]:
        pass

    @abc.abstractmethod
    def clear(self) -> None:
        pass

    @abc.abstractmethod
    def get_by_executable_path(self, executable_path: str) -> RegistryItem | None:
        pass


class MemoryRegistry(BaseRegistry):
    _functions: list[RegistryItem] = []

    def register_item(self, item: RegistryItem) -> None:
        self._functions.append(item)
        logger.info(
            "Registered function",
            function_name=item.function.__name__,
            executable_path=item.executable_path,
            args=item.args,
            backfilling_lookback=item.backfilling_lookback,
        )

    def get_functions(self) -> list[RegistryItem]:
        return self._functions

    def clear(self) -> None:
        self._functions = []

    def get_by_executable_path(self, executable_path: str) -> RegistryItem:
        for registry_item in self.get_functions():
            if registry_item.executable_path == executable_path:
                return registry_item
        # TODO: Improve this
        raise Exception("Function Not Found")


task_registry = MemoryRegistry()
